/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 * Pan - A Newsreader for Gtk+
 * Copyright (C) 2002  Charles Kerr <charles@rebelbase.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <config.h>

#include <errno.h>
#include <string.h>

#include <glib.h>
#include <gdk/gdkkeysyms.h>
#include <gtk/gtk.h>

#include <pan/base/argset.h>
#include <pan/base/debug.h>
#include <pan/base/file-grouplist.h>
#include <pan/base/pan-config.h>
#include <pan/base/pan-glib-extensions.h>
#include <pan/base/pan-i18n.h>
#include <pan/base/serverlist.h>

#include <pan/articlelist.h>
#include <pan/dialogs/dialogs.h>
#include <pan/globals.h>
#include <pan/grouplist.h>
#include <pan/group-action.h>
#include <pan/group-ui.h>
#include <pan/gui.h>
#include <pan/prefs.h>
#include <pan/queue.h>
#include <pan/task-grouplist.h>
#include <pan/util.h>

#include <pan/xpm/pan-pixbufs.h>

#include <pan/group-store.h>

/***
****
****  PRIVATE VARIABLES
****
***/

static GtkItemFactoryEntry group_popup_entries[] =
{
	/* h */ {N_("/Get New _Headers"), NULL, group_action_selected_download_new, 0, NULL},
	/* b */ {N_("/Get New Headers and _Bodies"), NULL, group_action_selected_download_new_and_bodies, 0, NULL},
	/*   */ {N_("/Get New Headers in Subscribed Groups"), NULL, group_action_subscribed_download_new, 0, NULL},
	/* o */ {N_("/More Download _Options..."), NULL, group_action_selected_download_dialog, 0, NULL},
	/* c */ {N_("/Refresh Article _Counts"), NULL, group_action_selected_update_count_info, 0, "<StockItem>", GTK_STOCK_REFRESH},
	/*   */ {N_("/---"), NULL, NULL, 0, "<Separator>"},
	/* s */ {N_("/_Subscribe"), NULL, group_action_selected_subscribe, 0, "<ImageItem>", icon_newsgroup},
	/* u */ {N_("/_Unsubscribe"), NULL, group_action_selected_unsubscribe, 0, NULL},
	/* p */ {N_("/Group _Properties..."), NULL, group_action_selected_properties, 0, "<StockItem>", GTK_STOCK_PROPERTIES},
	/*   */ {N_("/---"), NULL, NULL, 0, "<Separator>"},
	/* r */ {N_("/Mark Group _Read"), NULL, group_action_selected_mark_read, 0, NULL},
	/* d */ {N_("/_Delete Group's Articles"), NULL, group_action_selected_empty, 0, "<StockItem>", GTK_STOCK_DELETE},
	/*   */ {N_("/---"), NULL, NULL, 0, "<Separator>"},
	/* t */ {N_("/Dele_te Group"), NULL, group_action_selected_destroy, 0, "<StockItem>", GTK_STOCK_DELETE}
};

extern GtkWidget   * ttips;
static GtkItemFactory * popup_factory        = NULL;
static GtkWidget * popup_menu                = NULL;

static int           group_mode              = GROUP_ALL;
static Server      * my_server               = NULL;
static GtkWidget   * grouplist_mode_menu     = NULL;
static GtkWidget   * groupname_filter_entry  = NULL;
static char       * groupname_filter_string = NULL;

PanCallback * grouplist_group_selection_changed = NULL;


/***
****
****  GROUP SELECTION
****
***/

/***
**** id for the selection-changed callback
**/
static gulong select_id=0;

static void
grouplist_set_selected_groups_nolock(const GPtrArray * a)
{
	g_assert(select_id!=0);
	if G_UNLIKELY(a==NULL || a->len==0) return;

	debug_enter ("grouplist_set_selected_groups_nolock");
	
	GtkTreeView *view=GTK_TREE_VIEW(Pan.group_tree);
	GtkTreeSelection *sel=gtk_tree_view_get_selection(view);
	GtkTreeModel *model=gtk_tree_view_get_model(view);
	GtkTreeIter iter;
	
	g_signal_handler_block(sel,select_id);
	
	gtk_tree_selection_unselect_all(sel);
	if( gtk_tree_model_get_iter_first(model, &iter) )
	{
		Group *g;
		gint x;
		do
		{
			gtk_tree_model_get(model,&iter,GL_GROUP_COL,&g,-1);
			for(x=0;x<a->len;x++)
			{
				if(g_ptr_array_index(a,x)==g)
					gtk_tree_selection_select_iter(sel,&iter);
			}
		}while(gtk_tree_model_iter_next(model,&iter));
	}
	
	g_signal_handler_unblock(sel,select_id);
	debug_exit ("grouplist_set_selected_groups_nolock");
}

Group*
grouplist_get_selected_group(void)
{
	/* old code used pan_lock */
	debug_enter("grouplist_get_selected_group");
	Group *retval=NULL;
	GtkTreeSelection *sel=gtk_tree_view_get_selection(GTK_TREE_VIEW(Pan.group_tree));
	
	if(gtk_tree_selection_count_selected_rows(sel)>0)
	{
		GtkTreeModel *model;
		GList *list=gtk_tree_selection_get_selected_rows(sel,&model);
		GtkTreeIter iter;
		
		gtk_tree_model_get_iter(model,&iter,(GtkTreePath*)(list->data));
		gtk_tree_model_get(model,&iter,GL_GROUP_COL,&retval,-1);
		g_list_foreach(list,(GFunc)gtk_tree_path_free,NULL);
		g_list_free(list);
	}
	
	debug_exit ("grouplist_get_selected_group");
	return retval;
}

/***
**** utility function for grouplist_get_selected_groups
**** adds group from iter to GPtrArray
***/
static void gl_add_selected(GtkTreeModel *model, GtkTreePath *path,
	GtkTreeIter *iter, gpointer data)
{
	Group *group;
	gtk_tree_model_get(model,iter,GL_GROUP_COL,&group,-1);
	g_ptr_array_add((GPtrArray*)data, group);
}

GPtrArray* grouplist_get_selected_groups(void)
{
	debug_enter ("grouplist_get_selected_groups");
	
	GtkTreeSelection *sel=gtk_tree_view_get_selection(GTK_TREE_VIEW(Pan.group_tree));
	GPtrArray *retval;
	gint count=gtk_tree_selection_count_selected_rows(sel);
	
	if(count>0)
	{
		retval=g_ptr_array_sized_new(count);
		gtk_tree_selection_selected_foreach(sel,gl_add_selected,(gpointer)retval);
	}else
	{
		Group *group=articlelist_get_group();
		retval=g_ptr_array_new();
		if(group!=NULL)
			g_ptr_array_add( retval, group);
	}
	
	debug_exit ("grouplist_get_selected_groups");
	return retval;
}

static int
grouplist_select_all_idle(gpointer unused)
{
	/* old code used pan_lock around selection */
	
	GtkTreeSelection *sel=gtk_tree_view_get_selection(GTK_TREE_VIEW(Pan.group_tree));
	gtk_tree_selection_select_all(sel);
	
	return 0;
}
void
grouplist_select_all(void)
{
	gui_queue_add (grouplist_select_all_idle, NULL);
}

static int
grouplist_deselect_all_idle(gpointer unused)
{
	
	GtkTreeSelection *sel=gtk_tree_view_get_selection(GTK_TREE_VIEW(Pan.group_tree));
	gtk_tree_selection_unselect_all(sel);
	
	return 0;
}
void
grouplist_deselect_all(void)
{
	gui_queue_add (grouplist_deselect_all_idle, NULL);
}


/***
****
****  REFRESH GROUPLIST
****
***/

static void
grouplist_rebuild_nolock (GPtrArray * groups, GPtrArray * sel)
{
	debug_enter("grouplist_rebuild_nolock");
	
	GtkTreeView *view=GTK_TREE_VIEW(Pan.group_tree);
	GPatternSpec *pattern =NULL;
	const gboolean do_filter= is_nonempty_string(groupname_filter_string);
	GroupStore *store;
	int i;
	
	/* it may take a while to make the new list
	** so remove the old one first to prevent users
	** from trying to use it.
	*/
	if(groups->len>100)
	{
		store=GROUP_STORE(group_store_new(NULL));
		gtk_tree_view_set_model(view,GTK_TREE_MODEL(store));
		g_object_unref(store);
	}
	store=GROUP_STORE(group_store_new(groups));
	if(do_filter)
	{
		pattern = g_pattern_spec_new (groupname_filter_string);
		group_store_set_filter(store,pattern);
	}
	gtk_tree_view_set_model(view,GTK_TREE_MODEL(store));
	g_object_unref(store);
	/* refresh the selection */
	if (sel!=NULL && sel->len!=0)
		grouplist_set_selected_groups_nolock (sel);

	debug_exit("grouplist_rebuild_nolock");
}

static void grouplist_server_groups_added_cb (gpointer, gpointer, gpointer);

static int
grouplist_update_ui_mainthread (gpointer unused)
{
	char * title;
	GtkTreeView * view = GTK_TREE_VIEW(Pan.group_tree);
	GPtrArray * groups;
	GPtrArray * sel;
	debug_enter ("grouplist_update_ui_mainthread");

	/**
	***  Get the list of groups that we're going to add.
	***
	***  Note that we temporarily turn off the groups_added
	***  callback -- the groups may be loaded as a result of
	***  our call, and we don't want to update the clist twice.
	***  This is kind of an ugly hack.
	**/
	pan_callback_remove( server_get_groups_added_callback(),
	                     grouplist_server_groups_added_cb, NULL);

	if (group_mode == GROUP_SUBSCRIBED)
	{
		groups = my_server==NULL
			? g_ptr_array_new ()
			: server_get_groups (my_server, SERVER_GROUPS_SUBSCRIBED);
	}
	else if (group_mode == GROUP_ALL)
	{
		groups = my_server==NULL
			? g_ptr_array_new ()
			: server_get_groups (my_server, SERVER_GROUPS_ALL);
	}
	else if (group_mode == GROUP_FOLDERS)
	{
		Server * s = serverlist_get_named_server(&INTERNAL_SERVER_NAME);
		groups = s==NULL
			? g_ptr_array_new()
			: server_get_groups (s, SERVER_GROUPS_ALL);
	}
	else if (group_mode == GROUP_NEW)
	{
		groups = g_ptr_array_new ();
		if (my_server != NULL) {
			guint i;
			GPtrArray * a = server_get_groups (my_server, SERVER_GROUPS_ALL);
			for (i=0; i<a->len; ++i) {
				Group * g = GROUP(g_ptr_array_index(a,i));
				if (group_is_new (g))
					g_ptr_array_add (groups, g);
			}
			g_ptr_array_free (a, TRUE);
		}
	}
	else
	{
		pan_warn_if_reached ();
		groups = g_ptr_array_new ();
	}

	pan_callback_add( server_get_groups_added_callback(),
	                  grouplist_server_groups_added_cb, NULL);

	/* get the list of selected groups */
	sel = grouplist_get_selected_groups ();

	/* set the groups column title */
	switch (group_mode) 
	{
		case GROUP_SUBSCRIBED: title = _("Subscribed");         break;
		case GROUP_ALL:        title = _("All Groups");         break;
		case GROUP_NEW:        title = _("New Groups");         break;
		case GROUP_FOLDERS:    title = _("Folders");            break;
		default:               title = "Bug!";                  break;
	}

	/* update UI */
	pan_lock();
	gtk_tree_view_column_set_title( gtk_tree_view_get_column(view, 1), title);
	grouplist_rebuild_nolock(groups, sel);
	pan_unlock();

	/* cleanup */
	g_ptr_array_free(sel, TRUE);
	g_ptr_array_free(groups, TRUE);
	debug_exit("grouplist_update_ui_mainthread");
	return 0;
}

static Server *
grouplist_get_visible_server (void)
{
	return group_mode==GROUP_FOLDERS
		? serverlist_get_named_server(&INTERNAL_SERVER_NAME)
		: my_server;
}

void
grouplist_refresh_nolock (void)
{
	debug_enter("grouplist_refresh_nolock");
	Server * server = grouplist_get_visible_server();

	if (server!=NULL)
	{
		gui_queue_add(grouplist_update_ui_mainthread, NULL);
	}

	debug_exit("grouplist_refresh_nolock");
}

static void
grouplist_server_groups_added_cb (gpointer server, gpointer groups, gpointer foo)
{
	Server * visible;
	debug_enter ("grouplist_server_groups_added_cb");

	visible = grouplist_get_visible_server ();
	if (visible!=NULL && visible==SERVER(server))
		grouplist_refresh_nolock ();

	debug_exit ("grouplist_server_groups_added_cb");
}

/***
****
****  KEYBOARD / MOUSE INTERACTION
****
***/

/*
 * This tries to reconcile * the desired behavior with GtkCList's quirks.
 * If the user selects a range, we don't want to activate the group
 * because the user is selecting the range to perform a batch operation.
 * However the GtkCList selection signal fires once per row, so there's
 * no way to tell if the first callback is part of a range or not.
 * The workaround is to push the work off to an idle function to be
 * invoked after all the selection callbacks are done.  The idle function
 * fires the grouplist seleciton pan_callback.
 */

/* static int _button_click_count = -1;
static int _mb = -1;
static gboolean _modifiers = FALSE; */

/* FIXME: single click select doesn't work */
static void
grouplist_selection_changed_cb (gpointer call_object,
                                gpointer call_arg,
                                gpointer user_data)
{
	GPtrArray * groups = (GPtrArray*)call_arg;
//	const gboolean activate = !_modifiers && _mb==1 && _button_click_count>=(single_click_selects_groups ? 2 : 1);
	const gboolean activate = !single_click_selects_groups;
	debug_enter ("grouplist_selection_changed_cb");

	/* sanity clause */
	g_return_if_fail (groups!=NULL);

	/* do we activate this group? */
	if (groups->len==1 && activate)
	{
		gui_page_set (HEADERS_PANE);
		articlelist_set_group (GROUP(g_ptr_array_index(groups,0)));
	}

	/* reset the click count for next time... */
/* 	_button_click_count = -1;
	_modifiers = FALSE;
	_mb = -1; */

	debug_exit ("grouplist_selection_changed_cb");
}

//FIXME: this doesn't work right yet
// x & y are screen coordinates
// allocation is relative to parent
static void position_popup(GtkMenu *menu, gint *x, gint *y, gboolean *push_in, gpointer data)
{
/*	GtkWidget *widget=GTK_WIDGET(data);
	GdkScreen *screen = gtk_widget_get_screen (widget);
	gint monitor_num;
	GdkRectangle monitor;
	GtkRequisition req;
	
	gdk_window_get_origin(widget->window,x,y);
	monitor_num = gdk_screen_get_monitor_at_point (screen, *x, *y);
	gtk_menu_set_monitor (menu, monitor_num);
	gdk_screen_get_monitor_geometry (screen, monitor_num, &monitor);
	gtk_widget_size_request (GTK(WIDGET(menu), &req);

	*x = CLAMP (*x, monitor.x, monitor.x + MAX (0, monitor.width - req.width));
	*y = CLAMP (*y, monitor.y, monitor.y + MAX (0, monitor.height - req.height));
*/
*x=200;
*y=200;
	*push_in = FALSE;
}

static void do_popup_menu (GtkWidget *my_widget, GdkEventButton *event)
{
	GtkWidget *menu;
	int button, event_time;
	GtkItemFactory *gif;
	gboolean have_group;
	Group *group;
	debug_enter("do_popup_menu");
  
	gif = popup_factory;
	group = grouplist_get_selected_group ();
	have_group = group!=NULL;
	
	menu_set_sensitive (gif, "/Mark Group Read", have_group);
	menu_set_sensitive (gif, "/Delete Group's Articles", have_group);
	menu_set_sensitive (gif, "/Get New Headers", have_group);
	menu_set_sensitive (gif, "/Get New Headers and Bodies", have_group);
	menu_set_sensitive (gif, "/Get New Headers in Subscribed Groups", TRUE);
	menu_set_sensitive (gif, "/More Download Options...", have_group);
	menu_set_sensitive (gif, "/Refresh Article Counts", have_group);
	menu_set_sensitive (gif, "/Subscribe", have_group);
	menu_set_sensitive (gif, "/Unsubscribe", have_group);
	menu_set_sensitive (gif, "/Group Properties...", have_group);
	menu_set_sensitive (gif, "/Delete Group", have_group);
	if (event)
	{
		button = event->button;
		event_time = event->time;
		gtk_menu_popup (GTK_MENU(popup_menu), NULL, NULL, NULL, NULL, button, event_time);
	}
	else
	{
		button = 0;
		event_time = gtk_get_current_event_time ();
		gtk_menu_popup (GTK_MENU(popup_menu), NULL, NULL, position_popup, my_widget, button, event_time);
	}
	debug_exit("do_popup_menu");
}

static gboolean
grouplist_button_press (GtkWidget *widget, GdkEventButton * event)
{
	gboolean retval = FALSE;
	GdkModifierType modifiers = 0;
	debug_enter ("grouplist_button_press");

	gdk_event_get_state ((GdkEvent*)event, &modifiers);
//	_modifiers = modifiers & (GDK_SHIFT_MASK || GDK_CONTROL_MASK || GDK_MOD1_MASK || GDK_MOD2_MASK || GDK_MOD3_MASK || GDK_MOD4_MASK || GDK_MOD5_MASK);
//	_modifiers = modifiers & gtk_accelerator_get_default_mod_mask();

	switch (event->button)
	{
		case 1:
		case 2:
/* 			_mb = event->button;
			_button_click_count = event->type==GDK_2BUTTON_PRESS ? 2 : 1;
*/
			retval = FALSE;
			break;

		case 3:
			do_popup_menu(widget,event);
			retval = TRUE;
			break;
	}

	debug_exit ("grouplist_button_press");
	return retval;
}

/*
** row_activated requires a double click or pressing enter to be fired
** this should always set the group
*/
static gboolean row_activated_cb( GtkTreeView *view, GtkTreePath *path,
	GtkTreeViewColumn *col, gpointer data)
{
	debug_enter("row_activated_cb");
	Group *cur=articlelist_get_group();
	Group *g;
	GtkTreeIter iter;
	GtkTreeModel *model=gtk_tree_view_get_model(view);
	gtk_tree_model_get_iter(model,&iter,path);
	gtk_tree_model_get(model,&iter,GL_GROUP_COL,&g,-1);
	if(g!=cur)
	{
		gui_page_set (HEADERS_PANE);
		articlelist_set_group (g);
	}
	debug_exit("row_activated_cb");
	return FALSE;
}

static grouplist_popup_menu_handler(GtkWidget *widget)
{
	do_popup_menu(widget,NULL);
	return TRUE;
}

static gboolean select_callback_pending = FALSE;

static int
select_row_cb_mainthread (gpointer data)
{
	GPtrArray * groups;
	debug_enter("select_row_cb_mainthread");

	/* get the selected groups */
	groups = grouplist_get_selected_groups();

	/* make the callbacks */
	pan_callback_call(grouplist_group_selection_changed,
	                   Pan.group_tree, groups);
	select_callback_pending = FALSE;

	/* if the header pane's group is no longer selected, hide it.
	 * http://bugzilla.gnome.org/show_bug.cgi?id=102244 */
	if (1) 
	{
		Group * group = articlelist_get_group ();
		if (group != NULL) 
		{
			guint i;
			gboolean match;
			for (match=FALSE, i=0; !match && i<groups->len; ++i)
				if (g_ptr_array_index(groups,i) == group)
					match = TRUE;
			if (!match)
				articlelist_set_group(NULL);
		}
	}

	/* cleanup */
	g_ptr_array_free(groups, TRUE);
	debug_exit("select_row_cb_mainthread");
	return FALSE;
}

static void selection_changed_cb(GtkTreeSelection *ts, gpointer data)
{
	debug_enter ("selection_changed_cb");

	if (!select_callback_pending)
	{
		select_callback_pending = TRUE;
		gui_queue_add (select_row_cb_mainthread, NULL);
	}

	debug_exit ("selection_changed_cb");
}

/*****
****** ccc
*****/

void
grouplist_get_all (void)
{
	if (my_server == NULL)
		pan_error_dialog ("No server selected.");
	else
		queue_add (TASK(task_grouplist_new (my_server, GROUPLIST_ALL)));
}

void
grouplist_get_new (void)
{
	if (my_server == NULL)
		pan_error_dialog ("No server selected.");
	else
		queue_add (TASK(task_grouplist_new (my_server, GROUPLIST_NEW)));
}

/*****
******
*****/

void
grouplist_add_subscribed_to_selection_nolock (void)
{
	GtkTreeView * view = GTK_TREE_VIEW(Pan.group_tree);
	GtkTreeModel *model=gtk_tree_view_get_model(view);
	GtkTreeSelection *sel=gtk_tree_view_get_selection(view);
	GtkTreeIter iter;
	Group *group;

	g_signal_handler_block(sel,select_id);
	
	if(gtk_tree_model_get_iter_first(model,&iter))
		do
		{
			gtk_tree_model_get(model,&iter, GL_GROUP_COL, &group);
			if (group_is_subscribed(group))
				gtk_tree_selection_select_iter(sel,&iter);
		} while(gtk_tree_model_iter_next(model,&iter));
	
	g_signal_handler_unblock(sel,select_id);
}


/***
****  GROUPS REMOVED
***/

static int
grouplist_server_groups_removed_idle (gpointer data)
{
	GPtrArray * groups;
	debug_enter ("grouplist_server_groups_removed_idle");

	/* if the group's server matches the visible server... */
	groups = (GPtrArray*) data;
	if (groups->len && (GROUP(g_ptr_array_index(groups,0))->server == grouplist_get_visible_server()))
	{
		guint i;
		GtkTreeView * view = GTK_TREE_VIEW(Pan.group_tree);
		GtkTreeModel *model=gtk_tree_view_get_model(view);
//		GPtrArray *iters=g_ptr_array_sized_new(groups->len);
		

		/* remove the rows that match the groups that were removed. */
		grouplist_refresh_nolock();
/*		pan_lock();
 		for (i=0; i<groups->len; ++i) {
			GtkTreeIter iter;
			if(!gtk_tree_model_get_iter_first(model,&iter))
				break;
			do
			{
				Group *g;
				gtk_tree_model_get(model,&iter,GL_GROUP_COL,&g,-1);
				if(g==g_ptr_array_index(groups,i))
					g_ptr_array_add(iters,(gpointer)gtk_tree_iter_copy(&iter));
			} while(gtk_tree_model_iter_next(model,&iter));
		}
		while(iters->len !=0)
		{
			GtkTreeIter *iter=(GtkTreeIter*)g_ptr_array_remove_index_fast(iters,iters->len-1);
			gtk_list_store_remove(GTK_LIST_STORE(model), iter);
			gtk_tree_iter_free(iter);
		} 
		g_ptr_array_free(iters,TRUE);
		pan_unlock(); */
	}

	/* cleanup */
	g_ptr_array_free (groups, TRUE);
	debug_exit ("grouplist_server_groups_removed_idle");
	return 0;
}
static void
grouplist_server_groups_removed_cb (gpointer server, gpointer groups, gpointer foo)
{
	gui_queue_add (grouplist_server_groups_removed_idle, pan_g_ptr_array_dup((GPtrArray*)groups));
}

/***
****
****  FILTERING
****
***/

static void
grouplist_refresh_filter_nolock (void)
{
	char * new_filter;
        char * pch;

	pch  = gtk_editable_get_chars( GTK_EDITABLE(groupname_filter_entry), 0, -1);

	/* build the new filter string */
	new_filter = NULL;
	g_strstrip (pch);
	if (is_nonempty_string (pch))
		new_filter = strchr(pch,'*') ? g_strdup(pch) : g_strdup_printf("*%s*", pch);

	/* if it differs from the old filter string, refresh */
	GroupStore *store=GROUP_STORE(gtk_tree_view_get_model(GTK_TREE_VIEW(Pan.group_tree)));
	if (pan_strcmp(new_filter, groupname_filter_string)) {
		replace_gstr(&groupname_filter_string, g_strdup(new_filter));
		if(new_filter)
		{
			GPatternSpec *pattern = g_pattern_spec_new(groupname_filter_string);
			group_store_set_filter(store,pattern);
		}else
			group_store_set_filter(store,NULL);
	}

	/* cleanup */
	g_free(pch);
	g_free(new_filter);
}

static gboolean
grab_focus_idle (gpointer w)
{
	pan_lock();
	gtk_widget_grab_focus(GTK_WIDGET(w));
	pan_unlock();
	return FALSE;
}

static gboolean
grouplist_filter_changed_cb_nolock (void)
{
	grouplist_refresh_filter_nolock();

	gui_queue_add(grab_focus_idle, Pan.group_tree);

	return FALSE;
}

/***
****
****  VIEW MODE
****
***/

static void
grouplist_set_view_mode_nolock_impl (int unused)
{
	debug_enter("grouplist_set_view_mode_nolock_impl");

	GtkComboBox *combo=GTK_COMBO_BOX(grouplist_mode_menu);
	GtkTreeModel *store=gtk_combo_box_get_model(combo);
	GtkTreeIter iter;
	int view_mode;
	
	gtk_combo_box_get_active_iter(combo,&iter);
	gtk_tree_model_get(store,&iter,1,&view_mode,-1);
	group_mode=view_mode;

	debug_exit("grouplist_set_view_mode_nolock_impl");
}

static void
grouplist_set_view_mode_nolock (GtkWidget *widget, gpointer data)
{
	grouplist_set_view_mode_nolock_impl(GPOINTER_TO_INT(data));
	grouplist_refresh_nolock();
}


/***
****
****  SERVER CHANGED / ACTIVATED
****
***/

static void
grouplist_dialog_response_cb (GtkDialog * dialog, int response, gpointer user_data)
{
	if (response == GTK_RESPONSE_YES)
		queue_add (TASK(task_grouplist_new (SERVER(user_data), GROUPLIST_ALL)));
	gtk_widget_destroy (GTK_WIDGET(dialog));
}

static void
grouplist_set_server_nolock (Server * server)
{
	Server * old_server;
	debug_enter ("grouplist_set_server_nolock");

	/* find out what server we have, and what kind of groups it has */
	old_server = my_server;
	my_server = server;

	if (my_server==NULL)
	{
		if (grouplist_get_visible_server()==old_server)
			gtk_tree_view_set_model( GTK_TREE_VIEW(Pan.group_tree),
				GTK_TREE_MODEL(group_store_new(NULL)) );
	}
	else
	{
		int view_mode;
		GPtrArray * groups = NULL;
		const ServerGroupsType have_type = file_grouplist_exists (my_server);

		/* if we have subbed, always load those, otherwise load all */
		if (have_type & SERVER_GROUPS_SUBSCRIBED)
		{
			view_mode = GROUP_SUBSCRIBED;

			groups = my_server==NULL
				? g_ptr_array_new ()
				: server_get_groups (my_server, SERVER_GROUPS_SUBSCRIBED);
		}
		else if (have_type)
		{
			view_mode = GROUP_ALL;

			groups = my_server==NULL
				? g_ptr_array_new ()
				: server_get_groups (my_server, SERVER_GROUPS_ALL);
		}
		else
		{
			GtkWidget * w;

			view_mode = GROUP_ALL;
			groups = NULL;

			if (grouplist_get_visible_server()==old_server)
				grouplist_rebuild_nolock (NULL, NULL);

			w = gtk_message_dialog_new (GTK_WINDOW(Pan.window),
			                            GTK_DIALOG_DESTROY_WITH_PARENT,
			                            GTK_MESSAGE_QUESTION,
			                            GTK_BUTTONS_YES_NO,
			                            _("We don't have a list of groups for \"%*.*s\".\nShall we download one?"),
			                            server->name.len, server->name.len, server->name.str);
			gtk_dialog_set_default_response (GTK_DIALOG(w), GTK_RESPONSE_YES);
			g_signal_connect (w, "response", G_CALLBACK(grouplist_dialog_response_cb), server);
			gtk_widget_show_all (w);
		}

		/* update the UI */
		GtkComboBox *combo=GTK_COMBO_BOX(grouplist_mode_menu);
		/* the GROUP* defines are off by 1 from the index*/
		gtk_combo_box_set_active(combo,view_mode-1);
//		if (groups != NULL)
//			grouplist_rebuild_nolock (groups, NULL);

		/* cleanup */
		if (groups != NULL)
			g_ptr_array_free (groups, TRUE);
	}

	debug_exit ("grouplist_set_server_nolock");
}

static int
grouplist_server_activated_idle (gpointer server)
{
	pan_lock ();
	grouplist_set_server_nolock (SERVER(server));
	pan_unlock ();
	return 0;
}
static void
grouplist_server_activated_cb (gpointer server, gpointer unused1, gpointer unused2)
{
	gui_queue_add (grouplist_server_activated_idle, server);
}

/***
****
****  LIFE CYCLE - SHUTDOWN
****
***/

void
grouplist_shutdown_module (void)
{
	/* stop listening... */
	pan_callback_remove (serverlist_get_server_activated_callback(),
	                     grouplist_server_activated_cb, NULL);
	pan_callback_remove (grouplist_group_selection_changed,
	                     grouplist_selection_changed_cb, NULL);
	pan_callback_remove (server_get_groups_added_callback(),
	                     grouplist_server_groups_added_cb, NULL);
	pan_callback_remove (server_get_groups_removed_callback(),
	                     grouplist_server_groups_removed_cb, NULL);
	
//	g_object_unref(sub_pixbuf);
//	g_object_unref(folder_pixbuf);
}

/***
****
****  LIFE CYCLE - START UP
****
***/

extern GtkAccelGroup * _main_accel_group;

static gboolean
entry_focus_in_cb (GtkWidget * w, GdkEventKey * event, gpointer user_data)
{
	g_object_ref (_main_accel_group);
	gtk_window_remove_accel_group (GTK_WINDOW(Pan.window), _main_accel_group);
	return FALSE;
}
static gboolean
entry_focus_out_cb (GtkWidget * w, GdkEventKey * event, gpointer user_data)
{
	gtk_window_add_accel_group (GTK_WINDOW(Pan.window), _main_accel_group);
	g_object_unref (_main_accel_group);
	return FALSE;
}

typedef struct
{
	char * name;
	int mode;
}
GrouplistModeMenuStruct;

static GtkWidget*
grouplist_mode_menu_create (void)
{
	debug_enter("grouplist_mode_menu_create");
	GtkListStore *store;
	GtkWidget *combo;
	GtkCellRenderer *cell;
	GtkTreeIter iter;
	const int row_qty=4;
	int i, index=0;
	GrouplistModeMenuStruct foo[] = {
		{NULL, GROUP_ALL},
		{NULL, GROUP_SUBSCRIBED},
		{NULL, GROUP_NEW},
		{NULL, GROUP_FOLDERS}
	};
	foo[0].name = _("All Groups");
	foo[1].name = _("Subscribed");
	foo[2].name = _("New Groups");
	foo[3].name = _("Folders");
	
	store=gtk_list_store_new(2, G_TYPE_STRING, G_TYPE_INT);
	for(i=0;i<row_qty;++i)
	{
		gtk_list_store_append(store,&iter);
		gtk_list_store_set(store,&iter,0,foo[i].name,1,foo[i].mode,-1);
		if (group_mode == foo[i].mode)
			index = i;
	}
	combo=gtk_combo_box_new_with_model(GTK_TREE_MODEL(store));
	cell = gtk_cell_renderer_text_new ();
	gtk_cell_layout_pack_start (GTK_CELL_LAYOUT (combo), cell, TRUE);
	gtk_cell_layout_set_attributes (GTK_CELL_LAYOUT (combo), cell, "text", 0, NULL);
	g_signal_connect(combo, "changed", G_CALLBACK(grouplist_set_view_mode_nolock),NULL);
	gtk_widget_show_all(combo);
	gtk_combo_box_set_active(GTK_COMBO_BOX(combo), index);

	debug_exit("grouplist_mode_menu_create");

	return combo;
}


gpointer
grouplist_create (void)
{
	GtkWidget * w;
	GtkWidget * vbox;
	GtkWidget * hbox;
	GtkWidget * toolbar;
	GtkTreeView * view;
	char * titles[5];
	debug_enter ("grouplist_create");

	titles[0] = "";
	titles[1] = _("Groups");
	titles[2] = _("Unread");
	titles[3] = _("Total");
	titles[4] = _("Description");

	/* create callbacks */
	grouplist_group_selection_changed = pan_callback_new ();

	/* register callbacks */
	pan_callback_add (serverlist_get_server_activated_callback(),
	                  grouplist_server_activated_cb, NULL);
	pan_callback_add (grouplist_group_selection_changed,
	                  grouplist_selection_changed_cb, NULL);
	pan_callback_add (server_get_groups_added_callback(),
	                  grouplist_server_groups_added_cb, NULL);
	pan_callback_add (server_get_groups_removed_callback(),
	                  grouplist_server_groups_removed_cb, NULL);

	toolbar = gtk_toolbar_new ();
	gtk_toolbar_set_orientation (GTK_TOOLBAR(toolbar), GTK_ORIENTATION_HORIZONTAL);
	gtk_toolbar_set_style (GTK_TOOLBAR(toolbar), GTK_TOOLBAR_TEXT);

	vbox = gtk_vbox_new (FALSE, GUI_PAD_SMALL);
	gtk_container_set_border_width (GTK_CONTAINER(vbox), GUI_PAD_SMALL); 
	hbox = gtk_hbox_new (FALSE, GUI_PAD);

	Pan.group_tree = gtk_tree_view_new();
	view=GTK_TREE_VIEW(Pan.group_tree);
	GtkTreeModel *model=GTK_TREE_MODEL(group_store_new(NULL));
	gtk_tree_view_set_model(view,model);
	g_object_unref(model);
	GtkTreeViewColumn *col;
	GtkCellRenderer *cell;
	cell=gtk_cell_renderer_pixbuf_new();
	col=gtk_tree_view_column_new_with_attributes(titles[0], cell,
		"pixbuf", GL_PIXBUF_COL, NULL);
	g_object_set(col,"fixed-width",22,"min-width",22,"max-width",22,NULL);
	gtk_tree_view_column_set_resizable(col,FALSE);
	gtk_tree_view_column_set_sizing(col,GTK_TREE_VIEW_COLUMN_FIXED);
	gtk_tree_view_append_column(view,col);
	cell=gtk_cell_renderer_text_new();
	col=gtk_tree_view_column_new_with_attributes(titles[1], cell,
		"text", GL_NAME_COL, NULL);
	gtk_tree_view_append_column(view,col);
	gtk_tree_view_column_set_fixed_width(col,200);
	gtk_tree_view_column_set_sizing(col,GTK_TREE_VIEW_COLUMN_FIXED);
	gtk_tree_view_column_set_resizable(col,TRUE);
	cell=gtk_cell_renderer_text_new();
	col=gtk_tree_view_column_new_with_attributes(titles[2], cell,
		"text", GL_UNREAD_COL, NULL);
	g_object_set(cell,"xalign",1.0,NULL);
	gtk_tree_view_append_column(view,col);
	gtk_tree_view_column_set_fixed_width(col,50);
	gtk_tree_view_column_set_sizing(col,GTK_TREE_VIEW_COLUMN_FIXED);
	gtk_tree_view_column_set_resizable(col,TRUE);
	cell=gtk_cell_renderer_text_new();
	col=gtk_tree_view_column_new_with_attributes(titles[3], cell,
		"text", GL_QTY_COL, NULL);
	g_object_set(cell,"xalign",1.0,NULL);
	gtk_tree_view_column_set_fixed_width(col,50);
	gtk_tree_view_column_set_sizing(col,GTK_TREE_VIEW_COLUMN_FIXED);
	gtk_tree_view_column_set_resizable(col,TRUE);
	gtk_tree_view_append_column(view,col);
	cell=gtk_cell_renderer_text_new();
	col=gtk_tree_view_column_new_with_attributes(titles[4], cell,
		"text", GL_DESC_COL, NULL);
	gtk_tree_view_column_set_fixed_width(col,400);
	gtk_tree_view_append_column(view,col);
	grouplist_update_font ();
//#if GTK_MAJOR_VERSION==2 && GTK_MINOR_VERSION>=4
//	g_object_set(view,"fixed-height-mode",TRUE,NULL);
//#endif
	gtk_tree_selection_set_mode(gtk_tree_view_get_selection(view),GTK_SELECTION_MULTIPLE);
	
	w = gtk_label_new (_("Groups"));
	gtk_toolbar_append_widget(GTK_TOOLBAR(toolbar), w, NULL, NULL);
	gtk_toolbar_append_space (GTK_TOOLBAR(toolbar));
	
	grouplist_mode_menu = grouplist_mode_menu_create();
	gtk_box_pack_start (GTK_BOX(hbox), grouplist_mode_menu, FALSE, FALSE, 0);
	
	w = gtk_label_new_with_mnemonic (_("F_ind:"));
	gtk_box_pack_start (GTK_BOX(hbox), w, FALSE, FALSE, 0);
	groupname_filter_entry = gtk_entry_new ();
	gtk_label_set_mnemonic_widget (GTK_LABEL(w), groupname_filter_entry);
	w = groupname_filter_entry;
	g_signal_connect (w, "focus_in_event", G_CALLBACK(entry_focus_in_cb), w);
	g_signal_connect (w, "focus_out_event", G_CALLBACK(entry_focus_out_cb), w);
	g_signal_connect (w, "focus_out_event", G_CALLBACK(grouplist_filter_changed_cb_nolock), w);
	g_signal_connect (w, "activate", G_CALLBACK(grouplist_filter_changed_cb_nolock), w);
	gtk_tooltips_set_tip (GTK_TOOLTIPS(ttips), w,
		_("Type in a group search string and press ENTER.  "
		  "Wildcards are allowed."), NULL);

	gtk_widget_show_all (toolbar);
	gtk_box_pack_start (GTK_BOX(hbox), w, TRUE, TRUE, 0);
	gtk_box_pack_start (GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
	
	w = gtk_scrolled_window_new (NULL, NULL);

	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (w),
					GTK_POLICY_AUTOMATIC,
					GTK_POLICY_AUTOMATIC);
	gtk_container_add (GTK_CONTAINER(w), Pan.group_tree);
    gtk_box_pack_start (GTK_BOX (vbox), w, TRUE, TRUE, 0);

	GtkTreeSelection *sel=gtk_tree_view_get_selection(view);
	select_id=g_signal_connect(sel, "changed", G_CALLBACK(selection_changed_cb),NULL);
	g_signal_connect(Pan.group_tree, "button_press_event",
	                  G_CALLBACK(grouplist_button_press), NULL);
	g_signal_connect(Pan.group_tree, "popup_menu",
						G_CALLBACK(grouplist_popup_menu_handler), NULL);
	g_signal_connect(Pan.group_tree, "row-activated", G_CALLBACK(row_activated_cb), NULL);


	/* Create the Popup Menu while we are here. */
	popup_menu = menu_create_items (group_popup_entries,
	                                G_N_ELEMENTS (group_popup_entries),
	                                "<GroupView>",
	                                &popup_factory,
	                                NULL);


	/* bootstrap */
	grouplist_set_server_nolock (serverlist_get_active_server());

	gtk_widget_show_all (vbox);
	debug_exit ("grouplist_create");
	return vbox;
}


/***
****
****  ccc
****
***/

void
grouplist_update_font (void)
{
	pan_lock ();
	if (!group_pane_custom_font_enabled)
		gtk_widget_modify_font (Pan.group_tree, gtk_widget_get_default_style()->font_desc);
	else {
		PangoFontDescription * pfd = pango_font_description_from_string (group_pane_custom_font);
		gtk_widget_modify_font (Pan.group_tree, pfd);
		pango_font_description_free (pfd);
	}
	pan_unlock ();
}

/***
****
****  GROUP WALKING
****
***/

static void
grouplist_internal_select_row (GtkTreeIter *iter,
                               gboolean  activate)
{
	GtkTreeView* view = GTK_TREE_VIEW(Pan.group_tree);
	GtkTreeSelection *sel=gtk_tree_view_get_selection(view);

	g_signal_handler_block(sel,select_id);
	gtk_tree_selection_unselect_all(sel);
	g_signal_handler_unblock(sel,select_id);
	gtk_tree_selection_select_iter(sel,iter);
}

static gboolean
grouplist_select_row_if_unread (GtkTreeIter *iter,
                                gboolean activate)
{
	GtkTreeView* view = GTK_TREE_VIEW(Pan.group_tree);
	GtkTreeModel *model=gtk_tree_view_get_model(view);
	Group * g;
	gtk_tree_model_get(model,iter,GL_GROUP_COL,&g,-1);
	const gulong total = g->article_qty;
	const gulong read = g->article_read_qty;
	gboolean retval = FALSE;

	if (read < total) {
		grouplist_internal_select_row (iter, activate);
		retval = TRUE;
	}

	return retval;
}
static void
grouplist_next_unread_group (gboolean activate)
{
	GtkTreeView* view = GTK_TREE_VIEW(Pan.group_tree);
	GtkTreeModel *model=gtk_tree_view_get_model(view);
	GtkTreeIter iter;

	pan_lock ();

	if(gtk_tree_model_get_iter_first(model, &iter))
	{
		do
		{
			if(grouplist_select_row_if_unread(&iter,activate))
				break;
		} while(gtk_tree_model_iter_next(model, &iter));
	}

	pan_unlock ();
}
void
grouplist_select_next_unread_group (void)
{
	grouplist_next_unread_group (FALSE);
}
void
grouplist_activate_next_unread_group (void)
{
	grouplist_next_unread_group (TRUE);
}
static void
grouplist_next_group (gboolean activate)
{
	GtkTreeView* view = GTK_TREE_VIEW(Pan.group_tree);
	GtkTreeModel *model=gtk_tree_view_get_model(view);
	GtkTreeSelection *sel=gtk_tree_view_get_selection(view);
	GtkTreeIter iter;
	gboolean done=FALSE;
	
	pan_lock ();

	if(gtk_tree_selection_count_selected_rows(sel))
	{
		GList *list=gtk_tree_selection_get_selected_rows(sel,NULL);
		gtk_tree_model_get_iter(model,&iter,list->data);
		if(gtk_tree_model_iter_next(model,&iter))
			done=TRUE;
		g_list_foreach(list,(GFunc)gtk_tree_path_free,NULL);
		g_list_free(list);
	}
	if(!done)
	{
		if(gtk_tree_model_get_iter_first(model,&iter))
			done=TRUE;
	}
	if(done)
		grouplist_internal_select_row(&iter,activate);

	pan_unlock ();
}
void
grouplist_select_next_group (void)
{
	grouplist_next_group (FALSE);
}
void
grouplist_activate_next_group (void)
{
	grouplist_next_group (TRUE);
}
